package com.jpattern.orm.generator.reflection;

import java.sql.ResultSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import com.jpattern.orm.classmapper.IClassMapper;
import com.jpattern.orm.classmapper.IColumn;
import com.jpattern.orm.crud.AColumnValueGenerator;
import com.jpattern.orm.exception.OrmConfigurationException;
import com.jpattern.orm.exception.OrmReflectionException;
import com.jpattern.orm.generator.IOrmPersistor;
import com.jpattern.orm.generator.ResultSetMethodHelper;

/**
 * A persistor implementation based on reflection
 * @author Francesco Cina'
 *
 * Mar 24, 2012
 */
public class ReflectionOrmPersistor<T> implements IOrmPersistor<T> {

	//	private final ILogger logger = OrmLogger.getOrmLogger(this.getClass());
	private final IClassMapper<T> classMapper;
	private final Map<String, FieldManipulator> manipulators = new TreeMap<String, FieldManipulator>();
	private VersionManipulator versionManipulator = new NullVersionManipulator();
	private GeneratorManipulator generatorManipulator = new NullGeneratorManipulator();
	private final ResultSetMethodHelper resultSetMethodHelper;

	public ReflectionOrmPersistor(final IClassMapper<T> classMapper, final ResultSetMethodHelper resultSetMethodHelper) throws OrmConfigurationException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		this.classMapper = classMapper;
		this.resultSetMethodHelper = resultSetMethodHelper;
		this.initManipulators();
		this.initVersionManipulator();
		this.initGeneratorManipulator();
	}

	private void initManipulators() throws OrmConfigurationException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		final ReflectionManipulatorFactory reflectionManipulatorFactory = new ReflectionManipulatorFactory(this.resultSetMethodHelper);
		for (final String columnJavaName : this.classMapper.getAllColumnJavaNames()) {
			final IColumn column = this.classMapper.getColumnWithJavaName(columnJavaName);
			final AGetFieldManipulator getFieldManipulator = reflectionManipulatorFactory.getGetFieldManipulator(column);
			final ASetFieldManipulator setFieldManipulator = reflectionManipulatorFactory.getSetFieldManipulator(column);
			this.manipulators.put(columnJavaName, new FieldManipulator(getFieldManipulator, setFieldManipulator, column.getName()));
		}
	}

	private void initVersionManipulator() throws OrmConfigurationException, SecurityException, NoSuchMethodException {
		if (this.classMapper.getTableMap().isVersionable()){
			final String javaColumnName = this.classMapper.getTableMap().getVersionJavaFieldName();
			final IColumn column = this.classMapper.getColumnWithJavaName(javaColumnName);
			final FieldManipulator manipulator = this.manipulators.get(javaColumnName);
			this.versionManipulator = new ReflectionManipulatorFactory(this.resultSetMethodHelper).getVersionManipulator(column, manipulator);
		}
	}

	private void initGeneratorManipulator() throws OrmConfigurationException, SecurityException, NoSuchMethodException {
		if (this.classMapper.getTableMap().hasGeneratedField()) {
			final String javaColumnName = this.classMapper.getAllGeneratedColumnJavaNames()[0];
			final IColumn column = this.classMapper.getColumnWithJavaName(javaColumnName);
			final FieldManipulator fieldManipulator = this.manipulators.get(javaColumnName);
			final AColumnValueGenerator generator = this.classMapper.getTableMap().getGeneratorByJavaFieldName(javaColumnName);
			final long[] validValues = generator.getActivateValues();
			this.generatorManipulator = new ReflectionManipulatorFactory(this.resultSetMethodHelper).getGeneratorManipulator(column, fieldManipulator, validValues);
		}
	}

	@Override
	public T mapRow(final String rowNamePrefix, final ResultSet rs, final int rowNum) {
		final String[] allColumnNames = this.classMapper.getAllColumnJavaNames();
		try {
			final T entity = this.classMapper.getMappedClass().newInstance();
			for (final String columnJavaName : allColumnNames) {
				final FieldManipulator manipulator = this.manipulators.get(columnJavaName);
				manipulator.setValueFromResultSet(entity, rs, rowNamePrefix + manipulator.getColumnName());
			}
			return entity;
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	@Override
	public T clone(T entity) {
		try {
			final T entityCopy = this.classMapper.getMappedClass().newInstance();
			for ( final Entry<String, FieldManipulator> manipulatorEntry : manipulators.entrySet()) {
				final FieldManipulator manipulator = manipulatorEntry.getValue();
				manipulator.setValue(entityCopy, manipulator.getValue(entity));
			}
			return entityCopy;
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	@Override
	public Object[] allValues(final T entity) {
		return this.getValues(this.classMapper.getAllColumnJavaNames(), entity);
	}

	@Override
	public Object[] allNotGeneratedValues(final T entity) {
		return this.getValues(this.classMapper.getAllNotGeneratedColumnJavaNames(), entity);
	}

	@Override
	public Object[] primaryKeyValues(final T entity) {
		return this.getValues(this.classMapper.getPrimaryKeyColumnJavaNames(), entity);
	}

	@Override
	public Object[] notPrimaryKeyValues(final T entity) {
		return this.getValues(this.classMapper.getNotPrimaryKeyColumnJavaNames(), entity);
	}

	@Override
	public Object[] primaryKeyAndVersionValues(final T entity) {
		return this.getValues(this.classMapper.getPrimaryKeyAndVersionColumnJavaNames(), entity);
	}

	@Override
	public void increaseVersion(final T entity, final boolean firstVersionNumber) {
		if (this.classMapper.getTableMap().isVersionable()){
			try {
				this.versionManipulator.increaseVersion(entity, firstVersionNumber);
			} catch (final Exception e) {
				throw new OrmReflectionException(e);
			}
		}
	}

	@Override
	public void updateGeneratedValues(final ResultSet rs, final T entity) {
		final String[] allColumnNames = this.classMapper.getAllGeneratedColumnJavaNames();
		try {
			int i=1;
			for (final String columnJavaName : allColumnNames) {
				final FieldManipulator manipulator = this.manipulators.get(columnJavaName);
				manipulator.setValueFromResultSet(entity, rs, i);
				i++;
			}
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	private Object[] getValues(final String[] javaColumnNames, final T entity) {
		final Object[] result = new Object[javaColumnNames.length];
		try {
			for (int i=0; i<javaColumnNames.length ; i++) {
				final String javaColumnName = javaColumnNames[i];
				final FieldManipulator manipulator = this.manipulators.get(javaColumnName);
				result[i] = manipulator.getValue(entity);
			}
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
		return result;
	}

	@Override
	public boolean useKeyGenerators(final T entity) {
		try {
			return this.generatorManipulator.useGenerator(entity);
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	@Override
	public boolean hasGenerators() {
		try {
			return this.generatorManipulator.hasGenerator();
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	@Override
	public boolean hasConditionalGenerator() {
		try {
			return this.generatorManipulator.hasConditionalGenerator();
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

}
